import os
import sys
from setuptools import setup

MIN_PYTHON = (2, 7)
if sys.version_info < MIN_PYTHON:
    sys.stderr.write("Python {}.{} or later is required\n".format(*MIN_PYTHON))
    sys.exit(1)


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='morton-py',
    version='1.3',
    author='Andrew Kirilenko',
    author_email='iced@gojuno.com',
    maintainer='Alexander Verbitsky',
    maintainer_email='averbitsky@gojuno.com',
    description='Morton code pack/unpack library',
    long_description=read('README.rst'),
    keywords='z-order, morton coding, hashing',
    url='https://github.com/gojuno/morton-py',
    py_modules=['morton'],
    test_suite='test',
    license='BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'License :: OSI Approved :: BSD License',
    ],
)



import unittest
import time
import logging

from morton import Morton


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()


class TestMorton(unittest.TestCase):

    def test_packunpack(self):

        def do_test(dimensions, bits, cases):
            m = Morton(dimensions, bits)
            for values in cases:
                code = m.pack(*values)
                unpacked_values = m.unpack(code)
                assert unpacked_values == values, (unpacked_values, values)

        do_test(
            dimensions=2, bits=32,
            cases=[
                [1, 2],
                [2, 1],
                [(1 << 32)-1, (1 << 32)-1],
            ])
        do_test(
            dimensions=3, bits=21,
            cases=[
                [1, 2, 4],
                [4, 2, 1],
                [(1 << 21)-1, (1 << 21)-1, (1 << 21)-1],
                [1, 1, 1],
            ])
        do_test(
            dimensions=2, bits=1,
            cases=[
                [1, 1],
                [0, 0],
            ])
        do_test(
            dimensions=4, bits=16,
            cases=[
                [1, 2, 4, 8],
                [8, 4, 2, 1],
                [(1 << 16)-1, (1 << 16)-1, (1 << 16)-1, (1 << 16)-1],
                [1, 1, 1, 1],
            ])
        do_test(
            dimensions=6, bits=10,
            cases=[
                [1, 2, 4, 8, 16, 32],
                [32, 16, 8, 4, 2, 1],
                [1023, 1023, 1023, 1023, 1023, 1023],
            ])
        do_test(
            dimensions=64, bits=1,
            cases=[
                [1]*64,
            ])
        do_test(
            dimensions=2, bits=64,
            cases=[
                [1, 2],
                [2, 1],
                [(1 << 64)-1, (1 << 64)-1],
            ])

    def test_spacksunpack(self):

        def do_test(dimensions, bits, cases):
            m = Morton(dimensions, bits)
            ssize = (bits * dimensions) - 1
            low_range = (0 - (1 << ssize))
            up_range = ((1 << ssize) - 1)
            for values in cases:
                code = m.spack(*values)
                # check signed range
                assert code < up_range, (code, up_range, values)
                assert code > low_range, (code, low_range, values)
                unpacked_values = m.sunpack(code)
                assert unpacked_values == values, (unpacked_values, values)

        do_test(
            dimensions=2, bits=32,
            cases=[
                [1, 2],
                [2, 1],
                [(1<<31)-1, (1<<31)-1],
                [1, 1],
                [-1, -2],
                [-2, -1],
                [-((1 << 31) - 1), -((1 << 31) - 1)],
                [-1, -1],
            ])
        do_test(
            dimensions=4, bits=16,
            cases=[
                [1, 2, 4, 8],
                [8, 4, 2, 1],
                [(1<<15)-1, (1<<15)-1, (1<<15)-1, (1<<15)-1],
                [-1, -2, -4, -8],
                [-8, -4, -2, -1],
                [-((1 << 15) - 1), -((1 << 15) - 1), -((1 << 15) - 1), -((1 << 15) - 1)],
            ])


    def test_benchmark(self):
        m = Morton()
        t = time.time()
        for _ in range(100000):
            m.pack(3, 6)
        logger.info("time for pack: {}".format(time.time() - t))
        t = time.time()
        code = m.pack(3, 6)
        for _ in range(100000):
            m.unpack(code)
        logger.info("time for unpack: {}".format(time.time() - t))




class Morton(object):

    def __init__(self, dimensions=2, bits=32):
        assert dimensions > 0, 'dimensions should be greater than zero'
        assert bits > 0, 'bits should be greater than zero'

        def flp2(x):
            '''Greatest power of 2 less than or equal to x, branch-free.'''
            x |= x >> 1
            x |= x >> 2
            x |= x >> 4
            x |= x >> 8
            x |= x >> 16
            x |= x >> 32
            x -= x >> 1
            return x

        shift = flp2(dimensions * (bits - 1))

        masks = []
        lshifts = []
        max_value = (1 << (shift*bits))-1
        while shift > 0:
            mask = 0
            shifted = 0
            for bit in range(bits):
                distance = (dimensions * bit) - bit
                shifted |= shift & distance
                mask |= 1 << bit << (((shift - 1) ^ max_value) & distance)

            if shifted != 0:
                masks.append(mask)
                lshifts.append(shift)

            shift >>= 1

        self.dimensions = dimensions
        self.bits = bits
        self.lshifts = [0] + lshifts
        self.rshifts = lshifts + [0]
        self.masks = [(1 << bits) - 1] + masks
        self._size = self.bits * self.dimensions

    def __repr__(self):
        return '<Morton dimensions={}, bits={}>'.format(
            self.dimensions, self.bits)

    def split(self, value):
        # type: (int) -> int
        masks = self.masks
        lshifts = self.lshifts
        for o in range(len(masks)):
            value = (value | (value << lshifts[o])) & masks[o]
        return value

    def compact(self, code):
        # type: (int) -> int
        masks = self.masks
        rshifts = self.rshifts
        for o in range(len(masks)-1, -1, -1):
            code = (code | (code >> rshifts[o])) & masks[o]
        return code

    def shift_sign(self, value):
        # type: (int) -> int
        assert not(value >= (1<<(self.bits-1)) or value <= -(1<<(self.bits-1))), (value, self.bits)
        if value < 0:
            value = -value
            value |= 1 << (self.bits - 1)
        return value

    def unshift_sign(self, value):
        # type: (int) -> int
        sign = value & (1 << (self.bits - 1))
        value &= (1 << (self.bits - 1)) - 1
        if sign != 0:
            value = -value
        return value

    def pack(self, *args):
        # type: (List[int]) -> int
        assert len(args) <= self.dimensions
        assert all([(v < (1 << self.bits)) and (v >= 0) for v in args])

        code = 0
        for i in range(self.dimensions):
            code |= self.split(args[i]) << i
        return code

    def unpack(self, code):
        # type: (int) -> List[int]
        values = []
        for i in range(self.dimensions):
            values.append(self.compact(code >> i))
        return values

    def spack(self, *args):
        # type: (List[int]) -> int
        code = self.pack(*map(self.shift_sign, args))
        # convert from unsigned to signed
        return code if code < ((1 << self._size - 1) - 1) else code - (1 << self._size)

    def sunpack(self, code):
        # type: (int) -> List[int]
        values = self.unpack(code)
        return list(map(self.unshift_sign, values))

    def __eq__(self, other):
        return (
            self.dimensions == other.dimensions and
            self.bits == other.bits
        )
